import os
import re
import glob
import locale
import logging
import apt_pkg
from gettext import gettext as _
from SystemUpdater.Core.errors import *
from SystemUpdater.Core.enums import *
from SystemUpdater.Core.DistUpgradeCache import FreeSpaceRequired

def estimate_kernel_initrd_size_in_boot():
    """estimate the amount of space used by the kernel and initramfs in /boot,
    including a safety margin
    """
    kernel = 0
    initrd = 0
    kver = os.uname()[2]
    for f in glob.glob("/boot/*%s*" % kver):
        if f == '/boot/initrd.img-%s' % kver:
            initrd += os.path.getsize(f)
        # don't include in the estimate any files that are left behind by
        # an interrupted package manager run
        elif (f.find('initrd.img') >= 0 or f.find('.bak') >= 0
              or f.find('.dpkg-') >= 0):
            continue
        else:
            kernel += os.path.getsize(f)
    if kernel == 0:
        # logging.warning(
        #     "estimate_kernel_initrd_size_in_boot() returned '0' for kernel?")
        kernel = 16*1024*1024
    if initrd == 0:
        # logging.warning(
        #     "estimate_kernel_initrd_size_in_boot() returned '0' for initrd?")
        initrd = 100*1024*1024
    # add small safety buffer
    kernel += 1*1024*1024
    # safety buffer as a percentage of the existing initrd's size
    initrd_buffer = 1*1024*1024
    if initrd * 0.05 > initrd_buffer:
        initrd_buffer = initrd * 0.05
    initrd += initrd_buffer
    return kernel,initrd
KERNEL_SIZE, INITRD_SIZE = estimate_kernel_initrd_size_in_boot()

def kernel_spaces_requires(cache):
    # now calculate the space that is required on /boot
    # we do this by checking how many linux-image-$ver packages
    # are installed or going to be installed
    kernel_count = 0
    for pkg in cache:
        # we match against everything that looks like a kernel
        # and add space check to filter out metapackages
        if re.match("^linux-(image|image-debug)-[0-9.]*-.*", pkg.name):
            # upgrade because early in the release cycle the major version
            # may be the same or they might be -lts- kernels
            if pkg.marked_install or pkg.marked_upgrade:
                logging.debug("%s (new-install) added with %s to boot space" % (pkg.name, KERNEL_SIZE))
                kernel_count += 1
    # space calculated per LP: #1646222
    kernel_space = (kernel_count * KERNEL_SIZE
                        + (kernel_count + 1) * INITRD_SIZE)
    return kernel_space

def check_free_space(required_download=0,additional_required_space=0,kernel_space=0):
    """
    this checks if we have enough free space on /var, /boot and /usr
    with the given cache

    Note: this can not be fully accurate if there are multiple
            mountpoints for /usr, /var, /boot
    """

    class FreeSpace(object):
        " helper class that represents the free space on each mounted fs "
        def __init__(self, initialFree):
            self.free = initialFree
            self.need = 0

    def make_fs_id(d):
        """ return 'id' of a directory so that directories on the
            same filesystem get the same id (simply the mount_point)
        """
        for mount_point in mounted:
            if d.startswith(mount_point):
                return mount_point
        return "/"

    # this is all a bit complicated
    # 1) check what is mounted (in mounted)
    # 2) create FreeSpace objects for the dirs we are interested in
    #    (mnt_map)
    # 3) use the  mnt_map to check if we have enough free space and
    #    if not tell the user how much is missing
    mounted = []
    mnt_map = {}
    fs_free = {}
    try:
        with open("/proc/mounts") as mounts:
            for line in mounts:
                try:
                    (what, where, fs, options, a, b) = line.split()
                except ValueError as e:
                    logging.debug("line '%s' in /proc/mounts not understood (%s)" % (line, e))
                    continue
                if not where in mounted:
                    mounted.append(where)
        # make sure mounted is sorted by longest path
        mounted.sort(key=len, reverse=True)

        archivedir = apt_pkg.config.find_dir("Dir::Cache::archives")
        for d in ["/", "/usr", "/var", "/boot", archivedir, "/home", "/tmp/"]:
            d = os.path.realpath(d)
            fs_id = make_fs_id(d)
            if os.path.exists(d):
                st = os.statvfs(d)
                free = st.f_bavail * st.f_frsize
            else:
                logging.warning("directory '%s' does not exists" % d)
                free = 0
            if fs_id in mnt_map:
                logging.debug("Dir %s mounted on %s" %
                                (d, mnt_map[fs_id]))
                fs_free[d] = fs_free[mnt_map[fs_id]]
            else:
                logging.debug("Free space on %s: %s" %
                                (d, free))
                mnt_map[fs_id] = d
                fs_free[d] = FreeSpace(free)
        del mnt_map

        # sum up space requirements
        for (dir, size) in [(archivedir, required_download),
                            ("/usr", additional_required_space),
                            # plus 50M safety buffer in /usr
                            ("/usr", 50*1024*1024),
                            # ("/boot", kernel_space),
                            ("/tmp", 5*1024*1024),   # /tmp for dkms LP: #427035
                            ("/", 500*1024*1024),     # small safety buffer /
                            ]:
            # we are ensuring we have more than enough free space not less
            if size < 0:
                continue
            dir = os.path.realpath(dir)
            logging.debug("dir '%s' needs '%s' of '%s' (%f)" % (dir, size, fs_free[dir], fs_free[dir].free))
            fs_free[dir].free -= size
            fs_free[dir].need += size

        # check for space required violations
        required_list = {}
        for dir in fs_free:
            if fs_free[dir].free < 0:
                # ensure unicode here (LP: #1172740)
                free_at_least = apt_pkg.size_to_str(float(abs(fs_free[dir].free)+1))
                if isinstance(free_at_least, bytes):
                    free_at_least = free_at_least.decode(
                        locale.getpreferredencoding())
                free_needed = apt_pkg.size_to_str(fs_free[dir].need)
                if isinstance(free_needed, bytes):
                    free_needed = free_needed.decode(
                        locale.getpreferredencoding())
                # make_fs_id ensures we only get stuff on the same
                # mountpoint, so we report the requirements only once
                # per mountpoint
                required_list[make_fs_id(dir)] = FreeSpaceRequired(free_needed, make_fs_id(dir), free_at_least)
    except Exception as e:
        logging.error("free space check failed %s",str(e))
    
    if len(required_list) > 0:
        logging.error("Not enough free space: %s" % [str(i) for i in required_list])
        raise UpdateBaseError(ERROR_NOT_DISK_SPACE)
    else:
        logging.info("Check: Enough free space For system update...")
        
    return True