# UpdateManager.py
# -*- Mode: Python; indent-tabs-mode: nil; tab-width: 4; coding: utf-8 -*-
import os
import sys
import time
import json
import dbus
import shutil
import logging
import apt_pkg
import subprocess
import traceback
import collections
import hashlib
import random
import dbus.service
from .Core.errors import *
from .Core.enums import *
from datetime import date
from .Core.loop import mainloop
from gettext import gettext as _
from .Core.MyCache import MyCache
from .Core.PluginManager import *
from apt.debfile import DebPackage
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)
from .Core.UpdateList import UpdateList
from .Core.Database import Sqlite3Server
from aptdaemon.enums import ERROR_DAEMON_DIED
from .UpdatesAvailable import UpdatesAvailable
from .backend import InstallBackend,get_backend
from .Core.DataAcquisition import UpdateMsgCollector
from .Core.DeltaDownload import DeltaDownloadBackend
from SystemUpdater.Core.UpdateEssentialItem import UpdateSanityCheck,UpdateEssentialItem
from SystemUpdater.Core.PlymouthManager import PlymouthManager
from SystemUpdater.Core.UpdaterConfigParser import UpgradeConfig
from SystemUpdater.Core.shutdown_lock import InhibitShutdownLock,CatchShutdownAction
from SystemUpdater.Core.space_rules import check_free_space
from SystemUpdater.Core.GenericConfigParser import Gconfig
from SystemUpdater.Core.utils import deb_verify,PolicyKit_Authority,get_proc_from_dbus_name,get_dist,ApplicationSecInstall,is_offline_source
from .UpdateManagerDbus import UpdateManagerDbusController,UpdateManagerDbusControllerUtils,UPDATER_DBUS_INTERFACE,UPDATER_DBUS_PATH,UPDATER_DBUS_SERVICE
from aptdaemon.enums import ERROR_PACKAGE_MANAGER_FAILED

IMPORTANT_FIX_ENV = "allow-fix-env"
OPENKYLIN_DISTTRIBUTOR = "Openkylin"
BACKEND_PKG_NAME = 'kylin-system-updater'
FRONTEND_PKG_NAME = "kylin-update-frontend"
SYSTEM_UPDATE_GROUPS = "kylin-update-desktop-system"
GROUPS_PKG_NAME = 'kylin-update-desktop-config'
SOURCES_UPDATE_NAME = "kylin-software-properties"
APTD_PKG_NAME = "aptdaemon"
IMPORTANT_LIST_PATH = '/var/lib/kylin-software-properties/template/important.list'
SYSTEM_UPDATER_SOURCES = "/var/lib/kylin-system-updater/sources.list"

APT_PRIVATE_PATH = "/etc/kylin-system-updater/apt_private.conf"
APT_STANDARD_PATH = "/etc/kylin-system-updater/apt_standard.conf"

RESTART_APTD_SERVICE = "/tmp/restart-aptd"
# /tmp/restart-aptd

class UpdateManager():
    def __init__(self,options):
        self.options = options
        self.cache = None
        self.update_list = None
        self.last_install_mode = True

        self.now_working = InstallBackend.ACTION_DEFUALT_STATUS
        self.task_queue = collections.deque()

        self._check_exception_status()

        self.dist = get_dist()
        self.bus = dbus.SystemBus()
        
        #config and 数据库
        self.configs_uncover = UpgradeConfig(defaults_conf="system-updater-defaults.conf")
        self.configs_cover = UpgradeConfig(name = "system-updater-coverable.conf")
        self.uuconfigs = UpgradeConfig(datadir = "/var/lib/unattended-upgrades/", name = "unattended-upgrades-policy.conf")
        
        self.sqlite3_server = Sqlite3Server(self)
        self.collector = UpdateMsgCollector(self)

        #dbus
        self.dbusController = self._setup_dbus(self.bus)
        self.dbusControllerUtils = self._setup_dbus_utils(self.bus)

        #更新安装数据 
        if os.path.exists(UpdatesAvailable.UPDATE_AVAILABLE_DATA):
            local_data = UpdatesAvailable(self)
            local_data.read_local()
            if local_data.available:
                self.task_queue.append(local_data)
                try:
                    self.refresh_cache(True,True)
                except Exception as e:
                    logging.error(e)

        # 检查aptd升级完成之后是否需要重启服务
        if os.path.exists(RESTART_APTD_SERVICE):
            self.dbusController.make_aptdeamon_restart()
            time.sleep(1)

        # 其他配置 目前未使用
        self.apt_p2p_config = AptP2pConfigManager()
        self.delta = DeltaDownloadBackend(self)

    def run(self):
        """Start the daemon and listen for calls."""
        logging.info("Waiting for calls...")
        try:
            mainloop.run()
        except KeyboardInterrupt:
            self.dbusController.Quit(None)

    def _check_exception_status(self):
        logging.info("Check: Abonormal Device Startup conditions...")
        InhibitShutdownLock._check_unlock_sleep()

    def check_exception_fix(self):
        try:
            check_execp = InhibitShutdownLock()
            if check_execp.is_abnormal():
                time.sleep(2)
                # 标识上次更新失败
                self.refresh_cache(True,True)
                self._install_finished(success=False,error_code=ERROR_INSTALL_POWER_OUTAGE)
                self.start_available()
        except Exception as e:
            logging.error(e)
            tbk = traceback.format_exc()
            logging.error(str(tbk))
        
    @property
    def is_openkylin(self):
        return self.dist == OPENKYLIN_DISTTRIBUTOR

    #进行更新的操作
    def start_update(self,update_mode = InstallBackend.MODE_UPDATE_ALL, fix_env = True):
        try:
            self.now_working = InstallBackend.ACTION_UPDATE

            if self.options.check_network and not is_offline_source():
                self._check_network()

            if self.options.update_source and self.is_openkylin == False:
                def _update_important_success():
                    _fix_env = fix_env
                    if not self._get_important_data():
                        logging.info("Check: Server Push content is empty...")
                        self.dbusController.UpdateDetectFinishedNext(True,[],'')
                        self.now_working = InstallBackend.ACTION_DEFUALT_STATUS
                        return 
                    if update_mode == InstallBackend.MODE_UPDATE_ALL \
                                    and not os.path.exists(SYSTEM_UPDATER_SOURCES):
                        logging.info("Check: Souces(%s) file does not exits...",SYSTEM_UPDATER_SOURCES)
                        self.dbusController.UpdateDetectFinishedNext(True,[],'')
                        self.now_working = InstallBackend.ACTION_DEFUALT_STATUS
                    else:
                        self.start_update_cache(fix_env = _fix_env)

                self_uuid = self.configs_uncover.getWithDefault("statusForProperties","uuid",'',re_read=True)
                status = self.configs_uncover.getWithDefault("statusForProperties","last_install","",re_read=True)
                logging.info("Check: last update status:%s uuid:%s...",status,self_uuid)
                self._update_important(reply_handler=_update_important_success,error_handler=self._update_error_handler)
                return

            self.start_update_cache(update_mode,fix_env)
        except UpdateBaseError as excep:
            self._update_error_handler(excep.code)
        except UpdateProgressExit:
            pass
        except Exception as e:
            logging.error(e)
            tbk = traceback.format_exc()
            logging.error(str(tbk))
            self._update_error_handler(ERROR_PROGRAM_EXCEPTION)

    def start_update_cache(self,update_mode = InstallBackend.MODE_UPDATE_ALL,fix_env = True):
        #更新回调
        def _update_finished(parent,success,error_code):
            if success:
                if update_mode == InstallBackend.MODE_UPDATE_ALL:
                    self.start_available(fix_env)
                else:
                    self.dbusController.UpdateDetectFinishedNext(success,[''],error_code)
            else:
                if error_code == ERROR_DAEMON_DIED:
                    self.dbusController.UpdateDetectFinishedNext(success,[BACKEND_PKG_NAME],ERROR_DAEMON_DIED)
                elif error_code == ERROR_NOT_SOURCES_PATH:
                    self.dbusController.UpdateDetectFinishedNext(True,[],'')
                else:
                    self._update_error_handler(error_code)

        update_backend = get_backend(self, InstallBackend.ACTION_UPDATE,update_mode)
        update_backend.connect("finished", _update_finished)
        update_backend.start()

    def start_available(self,fix_env = True):
        try:
            # 刷新配置文件
            self.configs_cover.reRead()
            self.configs_uncover.reRead()
            self.uuconfigs.reRead()
            Gconfig(True)

            self.refresh_cache(True,True)

            self.update_list = UpdateList(self)

            self._check_iF_packages(self.cache)

            self._check_system_incomplete(self.cache,fix_env)
            
            self._check_system_broken(self.cache,fix_env)
            
            self._check_config_upgrade(self.cache,fix_env)

            self._check_self_upgrade(self.cache,fix_env)

            # ↑ ↑ ↑ ↑ 以上代码无法自我更新修复 不要随意改动添加

            self._check_other_upgrade(self.cache,fix_env)

            self._check_system_library()

            self._check_update_source()
            
            push_content = self._get_important_data()

            self.update_list.update_kylin(self.cache,push_content,self.is_openkylin,self._strategies_install_type())

            self._update_last_install(fix_env,push_content,self.update_list.get_push())

            # 自动更新和更新提示 都会来检查此处 进行到此处 执行更新前版本匹配检查
            if self.update_list.get_push():
                update_sanity = UpdateSanityCheck()
                success,desc = update_sanity.check()
                if not success:
                    # 控制面板 进行弹窗提示
                    if fix_env:
                        self.dbusController.ForceUpdatePopupRequestNext(True,self.update_list.get_push(),desc)
                    # 其他进程调用 直接发送失败
                    else:
                        self.dbusController.UpdateDetectFinishedNext(True,self.update_list.get_push(),'')
                else:
                    self.dbusController.UpdateDetectFinishedNext(True,self.update_list.get_push(),'')
            else:
                self.dbusController.UpdateDetectFinishedNext(True,self.update_list.get_push(),'')

            # 在低版本系统内 点关机和重启按钮情况触发关机安装，再次打开控制面板不清空下载
            if fix_env and self.task_queue:
                task = self.task_queue[-1]
                if task.signal_trigger and task.delay_handle:
                    task.delay_handle.unlock()
                self.task_queue.clear()

            self._print_now_sources()
            upload_thread = threading.Thread(target=self._upload_dpkg_status)
            upload_thread.start()
        except UpdateBaseError as excep:
            self._update_error_handler(excep.code)
        except UpdateProgressExit:
            pass
        except Exception as e:
            logging.error(e)
            tbk = traceback.format_exc()
            logging.error(str(tbk))
            self._update_error_handler(ERROR_PROGRAM_EXCEPTION)
    
    def start_available_next(self):
        if self.update_list:
            self.dbusController.UpdateDetectFinishedNext(True,self.update_list.get_push(),'')

    #更新下载前检查，包括 依赖计算+磁盘空间检查+重要软件包移除检查
    def start_resolver_depends(self,dist_upgrade = False,push_content = []):
        try:
            self.now_working = InstallBackend.ACTION_CHECK_RESOLVER

            self.refresh_cache(True,False)
            
            # 加载更新内容
            resolver = UpdatesAvailable(self)
            resolver.finial_update,resolver.upgrade_singles,resolver.upgrade_groups,resolver.upgrade_steps = self.update_list.classify_content(push_content)
            upgrade_group = resolver.upgrade_singles + resolver.upgrade_groups
            resolver.updating_data = self.update_list.copy_pkgs_data(resolver.upgrade_groups+resolver.upgrade_steps)

            remove_pkgs,remove_desc = resolver.start(self.cache,dist_upgrade)

            self.dbusController.UpdateDependResloveStatusNext(dist_upgrade,upgrade_group,remove_pkgs,remove_desc)

            self.task_queue.append(resolver)
        except UpdateBaseError as excep:
            self.dbusController.UpdateDependResloveStatusNext(dist_upgrade,upgrade_group,[],[],error_code=excep.code)
        except UpdateProgressExit as excep:
            pass
        except Exception as e:
            tbk = traceback.format_exc()
            logging.error(str(tbk))
        finally:
            self.now_working = InstallBackend.ACTION_DEFUALT_STATUS
    
    #进行下载的操作
    def start_download(self,upgrade_mode = InstallBackend.MODE_DEFAULT_STATUS,is_auto = False):
        #回调处理
        def _download_finished(parent,success,error_code):
            downloading_data = self.task_queue.pop()

            shutdown_install = downloading_data.shutdown_install
            downloading_content = downloading_data.upgrade_singles + downloading_data.upgrade_groups
            # 更新数据到本地
            if success:
                downloading_data.available = True
                self.task_queue.append(downloading_data)

            if shutdown_install:
                self.dbusController.UpdateDownloadFinishedNext(success,downloading_content,error_code)
                if success:
                    delay_shutdown = CatchShutdownAction()
                    if downloading_data.signal_trigger and delay_shutdown.lock():
                        delay_shutdown.connect_logind()
                        delay_shutdown.connect("trigger_install", self.start_poweroff)
                        downloading_data.delay_handle = delay_shutdown
            else:
                if is_auto:
                    # 运行时安装模式下 调用下载的接口 不能发送下载完成的信号会导致弹窗触发，所以切换到发送关机安装信号 仅针对自动更新 
                    self.dbusController.UpdateInstallFinished(success,downloading_content,\
                                    get_error_string_from_enum(error_code),get_error_description_from_enum(error_code))
                else:
                    if success:
                        self.start_run_install()
                    else:
                        self.dbusController.UpdateInstallFinishedNext(success,downloading_content,error_code)
        try:
            download_backend = get_backend(self,InstallBackend.ACTION_DOWNLOADONLY,upgrade_mode,self.task_queue[-1])
            download_backend.connect("finished", _download_finished)
            download_backend.start()
        except Exception as e:
            logging.error(e)

    #运行时安装
    def start_run_install(self):
        # 安装回调
        install_content = []
        def _install_finished(parent,success,error_code):
            install_data = self.task_queue.pop()
            install_content = install_data.get_content()

            self.dbusController.UpdateInstallFinishedNext(success,install_content,error_code)
            install_data.available = False
            if success:
                self.dbusController.CheckRebootRequired("self")
                # 弹出安装数据
                self.update_list.push_pop(install_content)
                
                # 全部更新完毕时刷新
                if not self.update_list.get_push():
                    self._refresh_uuid(success,error_code)
                    self._refresh_local_sources()
            else:
                self._refresh_uuid(success,error_code)
            
            self.refresh_cache(True,True)
            
        try:
            if not self.task_queue:
                raise UpdateBaseError(ERROR_NOT_UPGRADE_PACKAGES)

            install_data = self.task_queue[-1]
            install_content = install_data.get_content()

            self._check_pre_install(install_data)
            install_backend = get_backend(self,InstallBackend.ACTION_INSTALL,data = install_data)
            install_backend.connect("finished", _install_finished)
            install_backend.start()
        except UpdateBaseError as excep:
            self.dbusController.UpdateInstallFinishedNext(False,install_content,excep.code)
        except UpdateProgressExit as excep:
            pass
        except Exception as e:
            tbk = traceback.format_exc()
            logging.error(str(tbk))

    # 安装完成的回调
    def _install_finished(self,parent=None,success=True,error_code=''):
        try:
            if not self.task_queue:
                return 

            install_data = self.task_queue.pop()
            install_data.available = False
            install_content = install_data.get_content()
            if success:
                self._refresh_local_sources()

            self._record_finish_status(success,install_content,error_code)
            if install_data.finial_update:
                self._refresh_uuid(success,error_code)

            if not success and install_data.restorable and \
                                error_code == ERROR_PACKAGE_MANAGER_FAILED:
                self.configs_uncover.remain_conf()
                self.configs_cover.remain_conf()

            self.dbusController.UpdateInstallFinishedNext(success,install_content,error_code)
        except Exception as e:
            tbk = traceback.format_exc()
            logging.error(str(tbk)+str(e))
        finally:
            if error_code != ERROR_INSTALL_POWER_OUTAGE:
                CatchShutdownAction._unmask_lightdm_service()
                device_status = "reboot"
                plymouth = PlymouthManager()
                plymouth.show_countdown(success)
                if install_data:
                    device_status = install_data.device_status
                self._system_reboot_shutdown(device_status,self.options.prohibit_shutdown)

    #关机安装
    def start_poweroff(self,parent=None,state=None,dbus_trigger=False):
        try:
            install_data = None
            install_content = []
            if not bool(self.require_shutdown_install()):
                raise UpdateBaseError(ERROR_NOT_UPGRADE_PACKAGES)

            install_data = self.task_queue[-1]
            install_content = install_data.get_content()
            if state:
                install_data.device_status = state
            
            self._check_pre_install(install_data)

            install_backend = get_backend(self,InstallBackend.ACTION_INSTALL_SHUTDOWN,data=install_data)
            install_backend.connect("finished", self._install_finished)
            install_backend.start()
        except UpdateBaseError as excep:
            # dbus 触发时 弹窗提示用户充电，非dbus时记录错误并重启
            if dbus_trigger:
                self.dbusController.UpdateInstallFinishedNext(False,install_content,excep.code)
                # 再次抛出异常满足更新前端返回值
                raise UpdateBaseError(excep.code)
            else:
                self._install_finished(success=False,error_code=excep.code)
        except Exception as e:
            tbk = traceback.format_exc()
            logging.error(str(tbk)+str(e))

    def _check_pre_install(self,install_data):
        # 检查电量
        self._check_device_power(install_data.get_upgrade_num())
        
        check_free_space(additional_required_space=install_data.additional_required_space,\
                                            kernel_space=install_data.kernel_space)
        
        # 检查使用离线更新，但离线被卸载情况
        if install_data.offline_update and not is_offline_source():
            raise UpdateBaseError(ERROR_NO_OFFLINE_SOURCE)

    #进行修复破损的包的操作 apt install -f 
    def start_fix_broken(self):
        def _fix_finished(parent,success,error_code):
            if success:
                self.start_available()
            else:
                self._update_error_handler(error_code)

        fix_backend = get_backend(self, InstallBackend.ACTION_FIX_BROKEN)
        fix_backend.connect("finished", _fix_finished)
        fix_backend.start()

    #进行清空所有下载的文件
    def start_clean(self):
        clean_backend = get_backend(self, InstallBackend.ACTION_CLEAN)
        clean_backend.start()

    #进行 dpkg --configure
    def start_fix_incomplete(self):
        #安装回调
        def _fix_finished(parent,success,error_code):
            if success:
                self.start_available()
            else:
                self._update_error_handler(error_code)

        fix_backend = get_backend(self, InstallBackend.ACTION_FIX_INCOMPLETE)
        fix_backend.connect("finished", _fix_finished)
        fix_backend.start()

    def _check_network(self):
        self.dbusController.UpdateDetectStatusChanged(5,_("Checking network connection"))
        try:
            obj = self.bus.get_object("org.freedesktop.NetworkManager","/org/freedesktop/NetworkManager")
            interface = dbus.Interface(obj, "org.freedesktop.DBus.Properties")
            retval = interface.Get("org.freedesktop.NetworkManager","Connectivity",timeout=0.5)
        except Exception:
            retval = 4
        #1 表示没有网卡可以使用
        if retval == 1:
            raise UpdateBaseError(ERROR_NETWORK_FAILED)

    def _check_iF_packages(self,cache):
        # 对软件包的 iF状态进行修复 分两种情况 处理
        # 1、 检查到可升级时 先进行升级来修复，当升级完成还是没办法修复则进行第二步修复再次 
        # 2、当不可以升级时 强制删除安装前后脚本来修复系统环境
        need_upgrade = False
        self_upgrade = []
        failed_packages = []

        for pkg in cache:
            if pkg.is_installed and pkg._pkg.current_state == apt_pkg.CURSTATE_HALF_CONFIGURED:
                failed_packages.append(pkg.name)
        
        logging.info("Check: Packages state and Exception pkgs:%r...",failed_packages)

        for pkg_name in failed_packages:
            try:
                self_pkg = cache[pkg_name]
                self_pkg.mark_install(True,False,True)
                if self_pkg.marked_downgrade or self_pkg.marked_upgrade:
                    logging.info("Check: (%s) start fixing system From %s to %s...",pkg_name,\
                                    self_pkg.installed.source_version,self_pkg.candidate.source_version)
                    need_upgrade = True
                    self_upgrade.append(pkg_name)
                else:
                    self._remove_post_script(pkg_name)
            except SystemError:
                logging.error("Check: mark %s to upgrade Failed...",pkg_name)
                raise UpdateBaseError(ERROR_NOT_FIX_SYSTEM)
        
        if need_upgrade == True:
            self.dbusController.UpdateDetectStatusChanged(95,_("Repairing the system APT environment"))
            self.start_selfupgrade(self_upgrade)
            raise UpdateProgressExit()
    
    def _strategies_install_type(self):
        strategies_install = None
        if self.uuconfigs.getWithDefault("updateStrategiesManager","strategiesState",False):
            install_type = self.uuconfigs.getWithDefault("updateStrategiesManager","installType",DEFAULT_INSTALL_TYPE)
            if install_type != DEFAULT_INSTALL_TYPE:
                strategies_install = install_type
        return strategies_install


    def _remove_post_script(self,pkg_name):
        DPKG_SCRIPT_DIR = "/var/lib/dpkg/info"
        # 将脚本先进行临时存放 不删除
        TMP_SCRIPT_DIR = "/var/lib/kylin-system-updater/exception-scripts"
        script_suffix = [".postinst",".preinst"]

        for suf in script_suffix:
            script = os.path.join(DPKG_SCRIPT_DIR,pkg_name+suf)
            dest_scipt = os.path.join(TMP_SCRIPT_DIR,pkg_name+suf)
            if os.path.exists(script):
                if os.path.exists(dest_scipt):
                    os.remove(dest_scipt)
                logging.info("Exception Script(%s) exits to be removed to %s",script,TMP_SCRIPT_DIR)
                shutil.move(script,TMP_SCRIPT_DIR)

    #自我升级
    def start_selfupgrade(self,upgrade_content,fix_env=True):
        #自我升级回调 下面不要轻易改动 会引起严重问题
        def _selfupgrade_finished(parent,success,error_code):
            if success:
                channel_config = Gconfig().get("update_channel_upgrade",{})
                # restart_list = channel_config.get("restart_service_list",[APTD_PKG_NAME,BACKEND_PKG_NAME])
                notify_list = channel_config.get("restart_panel_list",[FRONTEND_PKG_NAME])
                if APTD_PKG_NAME in upgrade_content and BACKEND_PKG_NAME not in upgrade_content:
                    self.dbusController.make_aptdeamon_restart()
                if FRONTEND_PKG_NAME in upgrade_content and FRONTEND_PKG_NAME in notify_list:
                    self.dbusController.UpdateDetectFinishedNext(False,[BACKEND_PKG_NAME],PRIORITY_UPGRADE_SUCCCESSED)
                if SOURCES_UPDATE_NAME in upgrade_content:
                    self.dbusController.make_properties_restart()
                if BACKEND_PKG_NAME in upgrade_content:
                    if FRONTEND_PKG_NAME not in upgrade_content:
                        self.configs_uncover.setValue("SystemStatus","backend_restart",str(True))
                        if SOURCES_UPDATE_NAME in upgrade_content:
                            self.configs_uncover.setValue("SystemStatus","source_restart",str(True))
                    self.dbusController.Quit(None)
                    return
                if FRONTEND_PKG_NAME not in upgrade_content:
                    if SOURCES_UPDATE_NAME in upgrade_content:
                        self.start_update(fix_env=fix_env)
                    else:
                        self.start_available(fix_env=fix_env)
            else:
                self._update_error_handler(error_code)

        selfupgrade = get_backend(self,InstallBackend.ACTION_UPDATE,InstallBackend.MODE_SLEF_UPGRADE)
        selfupgrade.connect("finished", _selfupgrade_finished)
        selfupgrade.upgrade_content = upgrade_content
        selfupgrade.start()

    def _check_config_upgrade(self,cache,fix_env):
        self_upgrade = []

        #config包
        for pkg_name in [GROUPS_PKG_NAME]:
            if pkg_name in cache:
                self_pkg = cache[pkg_name]
                if self_pkg.is_installed:
                    if self_pkg.is_upgradable:
                        logging.info("Check: (%s) start upgrading From %s to %s...",pkg_name,\
                                        self_pkg.installed.version,self_pkg.candidate.version)
                        self_upgrade.append(pkg_name)
                    else:
                        # 已安装版本
                        cur_version = getattr(self_pkg.installed, "version", None)
                        sources_version = None
                        # 源中最高版本
                        for version_obj in self_pkg.versions:
                            if version_obj.downloadable == False:
                                continue
                            sources_version = version_obj.version
                            self_pkg.candidate = version_obj
                            break

                        if cur_version and sources_version:
                            # 当前版本大于源中的版本 将config包进行降级 
                            if apt_pkg.version_compare(cur_version,sources_version) > 0:
                                logging.info("Check: (%s) need to downgrade from %s to %s...",pkg_name,cur_version,sources_version)
                                self_upgrade.append(pkg_name+"="+sources_version)
                        logging.info("Check: (%s:%s) No need to upgrade...",pkg_name,self_pkg.installed.version)
                else:
                    logging.info("Check: (%s:%s) start new installing...",pkg_name,self_pkg.candidate.version)
                    self_upgrade.append(pkg_name)
            else:
                logging.error("Check: groups JSON ConfigPkgs(%s) is not in Cache...",pkg_name)
                raise UpdateBaseError(ERROR_NOT_GROUPS_CONFIG)

        if self_upgrade:
            try:
                cache.make_resolver_self_upgrade(self_upgrade)
            except Exception:
                raise UpdateBaseError(ERROR_NOT_SELFPKG_DEPENDENCIES)
            self.dbusController.UpdateDetectStatusChanged(95,_("Priority Upgrade Package being updated"))
            self.start_selfupgrade(self_upgrade,fix_env)
            raise UpdateProgressExit()


    def _check_self_upgrade(self,cache,fix_env):
        self_upgrade = []
        default_list = [BACKEND_PKG_NAME,APTD_PKG_NAME,FRONTEND_PKG_NAME,SOURCES_UPDATE_NAME]

        channel_config = Gconfig().get("update_channel_upgrade",None)
        if channel_config == None:
            logging.warning("Generic Config: update_channel_upgrade item is None...")
            upgrade_list =  default_list
        else:
            upgrade_list = channel_config.get("upgrade_list",default_list)

        for pkg_name in upgrade_list:
            if pkg_name in cache:
                self_pkg = cache[pkg_name]
                if self_pkg.is_installed:
                    if self_pkg.is_upgradable:
                        logging.info("Check: (%s) will upgrading From %s to %s...",pkg_name,\
                                        self_pkg.installed.source_version,self_pkg.candidate.source_version)
                        self_upgrade.append(pkg_name)
                    else:
                        logging.info("Check: (%s:%s) No need to upgrade...",pkg_name,self_pkg.installed.source_version)
                else:
                    logging.info("Check: (%s:%s) start new installing...",pkg_name,self_pkg.candidate.version)
                    self_upgrade.append(pkg_name)
            else:
                logging.info("Check: (%s) The upgrade package is not in Cache...",pkg_name)

        if self_upgrade and fix_env:
            try:
                cache.make_resolver_self_upgrade(self_upgrade)
            except Exception:
                raise UpdateBaseError(ERROR_NOT_SELFPKG_DEPENDENCIES)
            self.dbusController.UpdateDetectStatusChanged(95,_("Priority Upgrade Package being updated"))
            self.start_selfupgrade(self_upgrade,fix_env)
            raise UpdateProgressExit()

    def _check_other_upgrade(self,cache,fix_env):
        self_upgrade = []
        
        channel_config = Gconfig().get("update_other_upgrade",None)
        if channel_config == None:
            logging.warning("Generic Config: update_other_upgrade item is None...")
            upgrade_list =  []
            install_list =  []
        else:
            upgrade_list = channel_config.get("upgrade_list",[])
            install_list =  channel_config.get("install_list",[])

        for pkg_name in upgrade_list:
            if pkg_name in cache:
                self_pkg = cache[pkg_name]
                if self_pkg.is_installed:
                    if self_pkg.is_upgradable:
                        logging.info("Check: (%s) will upgrading From %s to %s...",pkg_name,\
                                        self_pkg.installed.source_version,self_pkg.candidate.source_version)
                        self_upgrade.append(pkg_name)
                    else:
                        logging.info("Check: (%s:%s) No need to upgrade...",pkg_name,self_pkg.installed.source_version)
                else:
                    logging.info("Check: (%s) Not to be installed...",pkg_name)
            else:
                logging.info("Check: (%s) The upgrade package is not in Cache...",pkg_name)

        for pkg_name in install_list:
            if pkg_name in cache:
                self_pkg = cache[pkg_name]
                if not self_pkg.is_installed:
                    logging.info("Check: (%s:%s) start new installing...",pkg_name,self_pkg.candidate.version)
                    self_upgrade.append(pkg_name)
            else:
                logging.info("Check: (%s) The upgrade package is not in Cache...",pkg_name)

        if self_upgrade and fix_env:
            try:
                cache.make_resolver_self_upgrade(self_upgrade)
            except Exception:
                raise UpdateBaseError(ERROR_NOT_SELFPKG_DEPENDENCIES)
            self.dbusController.UpdateDetectStatusChanged(95,_("Priority Upgrade Package being updated"))
            self.start_selfupgrade(self_upgrade,fix_env)
            raise UpdateProgressExit()

    def _check_device_power(self,pkgs_mun):
        def _battry_limit(power_type,pkgs_mun):
            # 最小电量限制
            MINI_POWER = 20
            # 最大电量限制
            MAX_POWER = 50

            # 最大软件包数量阈值
            MAX_PKGS_MUN = 500

            # 插入电源的标识为   插上电源充电 1 或者 插上电源充电，充满4 
            plug_ac = (power_type == 1 or power_type == 4)
            if not plug_ac:
                return 50

            # 基准数
            radix = (MAX_POWER - MINI_POWER) / MAX_PKGS_MUN

            # 根据软件包数量 计算电量要求
            requires = int(MINI_POWER + radix * pkgs_mun)
            if requires > MAX_POWER:
                requires = MAX_POWER
            return requires

        power_percentage = 0
        requires = 0
        try:
            # 支持调试模式 进行关闭
            if not self.configs_cover.getWithDefault("TestsConfig","check_device_power",str(True)):
                return 
            obj = self.bus.get_object("org.freedesktop.UPower","/org/freedesktop/UPower/devices/DisplayDevice")
            interface = dbus.Interface(obj, "org.freedesktop.DBus.Properties")

            # 获取供电方式
            power_type = int(interface.Get("org.freedesktop.UPower.Device","State",timeout=0.5))
            
            logging.info("Check: Currently powered state:%d",power_type)
            # 非电池方式直接退出检查
            # 1. 台式机为0   2. 插上电源充电 1   3. 未插入电源，用电池2 4. 插上电源充电，充满4 
            # if power_type == 0 or power_type == 1 or power_type == 4:
            if power_type == 2:
                # 电池供电方式
                power_percentage = int(interface.Get("org.freedesktop.UPower.Device","Percentage",timeout=0.5))
                requires = _battry_limit(power_type,pkgs_mun)
                logging.info("Check: Now device power is percentage:%d and requires:%d...",power_percentage,requires)

        except Exception as e:
            logging.error(str(e))

        if power_percentage != 0 and power_percentage < requires:
            logging.error("Not enough device power to shutdown install")
            raise UpdateBaseError(ERROR_DEVICE_LOW_BATTERY)  
        return 
        
    def _check_system_incomplete(self,cache,fix_env):
        if cache.dpkgJournalDirty() == True:
            if fix_env:
                self.start_fix_incomplete()
                raise UpdateProgressExit()
            else:
                raise UpdateBaseError(ERROR_NOT_FIX_SYSTEM)

    def _check_system_broken(self,cache,fix_env):
        if cache.is_system_broken() == True:
            if fix_env:
                _,_,pkgs_remove,_ = cache.fix_broken_for_update()
                if pkgs_remove:
                    # apt软件包环境修复，需要删除软件包时检查移除白名单检查，当移除的软件包在白名单则允许删除，不再弹窗
                    update_check = UpdateEssentialItem()
                    update_check.check_white(pkgs_remove)
                    if pkgs_remove:
                        self.dbusController.UpdateFixBrokenStatus(True,True,pkgs_remove,[''],[''],'','')
                        self.dbusController.UpdateDetectFinishedNext(False,[""],'')
                    else:
                        self.start_fix_broken()
                else:
                    self.start_fix_broken()
                raise UpdateProgressExit()
            else:
                raise UpdateBaseError(ERROR_NOT_FIX_SYSTEM)
    
    def _check_system_library(self):
        illegal_library = []

        skip_check_list = Gconfig().get("skip_check_list",[])
        if ERROR_NOT_NECESSARY_LIB in skip_check_list:
            return 
        
        try:
            system_library = []
            for item in os.listdir("/usr/local/lib"):
                if os.path.isfile(os.path.join("/usr/local/lib", item)):
                    system_library.append(item)

            illegal_library = self._check_necessary_library(system_library)
        except Exception as exc:
            logging.warning(str(exc))

        if len(illegal_library) != 0:
            logging.info("Check: illegal library: %r ...", illegal_library)
            raise UpdateBaseError(ERROR_NOT_NECESSARY_LIB)
        else:
            logging.info("Check: not illegal library...")

    def _check_update_source(self):
        skip_check_list = Gconfig().get("skip_check_list",[])
        if SYSTEM_UPDATER_SOURCES in skip_check_list:
            return 

        try:
            file_access = os.access(SYSTEM_UPDATER_SOURCES, os.W_OK)
        except Exception as exc:
            logging.error("Access update source file failed")
            logging.error(str(exc))
            raise UpdateBaseError(ERROR_SOURCE_WRITE_DENIED)
        else:
            if file_access is not True:
                raise UpdateBaseError(ERROR_SOURCE_WRITE_DENIED)
            else:
                pass

    def _check_necessary_library(self,system_library):
        illegal_library = []
        local_library = {}

        for sl in system_library:
            if ".so." in sl or sl.endswith(".so"):
                if not os.path.islink(sl):
                    process = subprocess.Popen("dpkg -S " + os.path.basename(sl), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    output, error = process.communicate()
                    if not error:
                        install_local, library_legal, library_path = self._check_library_legal(sl, output.decode('utf-8'))
                        if not install_local and not library_legal:
                            illegal_library.append(library_path)
                            local_library[str(sl)] = str(library_path)
                    else:
                        logging.warning("dpkg -S result: %s .",str(error.decode('utf-8')).strip())
                else:
                    logging.debug("'%s' is link file, should ignore.", sl)

        logging.debug("Check: library: %s .", local_library)
        return illegal_library

    def _check_library_legal(self, system_library, output):
        is_library_legal = True
        library_legal = ''
        lines = output.splitlines()

        for line in lines:
            if ": " in line:
                lib_path = line.split(": ")[1].strip()

                if lib_path == system_library:
                    logging.debug("'%s' is install to local, should ignore.", lib_path)
                    return True,is_library_legal,''

                if os.path.basename(system_library) == os.path.basename(line) and lib_path.startswith("/usr/lib/"):
                    logging.debug("'%s' is legal library.", lib_path)
                    is_library_legal = False
                    library_legal = line
        
        return False,is_library_legal,library_legal

    #更新important.list的本次升级的列表
    def _update_important(self,reply_handler,error_handler):
        def _update_important_reply(retval):
            if bool(retval) == False:
                error_handler(ERROR_UPDATE_SOURCE_FAILED)
            else:
                reply_handler()
        def _update_important_error(retval):
            logging.error(str(retval))
            error_handler(ERROR_UPDATE_SOURCE_FAILED)
        
        try:
            obj = self.bus.get_object('com.kylin.software.properties', '/com/kylin/software/properties')

            methods = dbus.Interface(obj, dbus_interface='org.freedesktop.DBus.Introspectable').Introspect()
            interface = dbus.Interface(obj, dbus_interface='com.kylin.software.properties.interface')
            if "checkUpdatesInStages" in methods:
                self.dbusController.UpdateDetectStatusChanged(10,_("Updating Source Template")+"(checkUpdatesInStages)")
                interface.checkUpdatesInStages(timeout=20,reply_handler=_update_important_reply,error_handler=_update_important_error)
            else:
                self.dbusController.UpdateDetectStatusChanged(10,_("Updating Source Template")+"(updateSourceTemplate)")
                interface.updateSourceTemplate(timeout=20,reply_handler=_update_important_reply,error_handler=_update_important_error)
        except Exception as e:
            logging.warning(str(e))
            raise UpdateBaseError(ERROR_UPDATE_SOURCE_FAILED)

    def _refresh_uuid(self,success,error_code):
        # 更新完成之后 设置上次安装状态
        if success:
            self.configs_uncover.setValue("statusForProperties","last_install",INSTALL_SUCCESS_FLAG)
        else:
            self.configs_uncover.setValue("statusForProperties","last_install",get_error_num_from_enum(error_code))
        
    def _record_finish_status(self,success,install_content,error_code):
        if success:
            self.configs_cover.setValue("statusForFinishPrompt","finished_status",INSTALL_SUCCESS_FLAG)
            self.configs_uncover.setValue("LastUpdateInfo","finished_status",INSTALL_SUCCESS_FLAG)
        else:
            self.configs_cover.setValue("statusForFinishPrompt","finished_status",get_error_num_from_enum(error_code))
            self.configs_uncover.setValue("LastUpdateInfo","finished_status",get_error_num_from_enum(error_code))

        self.configs_uncover.setListValue("LastUpdateInfo","update_content",install_content)

    def _update_last_install(self,fix_env,push_content,upgradeable_push):
        try:
            last_status = self.configs_uncover.getWithDefault("LastUpdateInfo","finished_status",INSTALL_SUCCESS_FLAG)

            # 获取源管理器的uuid
            tmpconfig = UpgradeConfig(datadir = "/var/lib/kylin-software-properties/config/", name = "updateID.conf")
            source_uuid = tmpconfig.getWithDefault("update", "uuid", "")
            # 当前系统更新保存的uuid
            self_uuid = self.configs_uncover.getWithDefault("statusForProperties","uuid",'')

            # 检查更新调用源更新管理的刷新源接口之后 
            # 当源服务的uuid不为空，并且和当前保存的uuid不一致时进行刷新，设置last_install="" 和 uuid=源更新管理uuid
            if source_uuid != "" and self_uuid != source_uuid:
                self.configs_uncover.setValue("statusForProperties","last_install","")
                self.configs_uncover.setValue("statusForProperties","uuid",source_uuid)

            # 推送内容或者可升级内容都是空时直接刷新状态 将上次安装设置为成功
            # 1. 系统真正升级成功了改为success 2. uuid不为空，检测出来是最新改为success。
            if upgradeable_push == [] and fix_env and self_uuid != "":
                # logging.info("upgradeable_push:%r %s...",upgradeable_push,str(fix_env))
                self.configs_uncover.setValue("statusForProperties","last_install",INSTALL_SUCCESS_FLAG)

            if upgradeable_push == [] and fix_env:
                logging.info("The software on this computer is up to date, try to refresh system version ...")
                self.sqlite3_server._repair_system_version()

            # 上次安装成功 不做任何处理
            if last_status == INSTALL_SUCCESS_FLAG or last_status == '':
                pass
            else:
                last_content = self.configs_uncover.getlist("LastUpdateInfo","update_content")
                
                # 比对版本号不同或者更新内容不相同 则认为认为存在新的推送
                # 推送更新的内容与上次不同
                # 系统为最新时也刷新
                if last_content != push_content or push_content != [] and upgradeable_push == []:
                    self.configs_uncover.setValue("LastUpdateInfo","finished_status",INSTALL_SUCCESS_FLAG)
                    self.configs_uncover.setListValue("LastUpdateInfo","update_content",[])

            # 适配前端多阶段更新 弹窗控制面板
            check_ignore = "/var/cache/kylin-update-manager/ignoreOrDelay"
            if os.path.exists(check_ignore):
                with open(check_ignore,'r') as f:
                    data = f.read()
                    content = data.split()
                    if content != "2107":
                        with open(check_ignore, 'w') as file:
                            file.write("2107")

        except Exception as e:
            logging.error(str(e))
    #更新完成刷新本地源信息
    def _refresh_local_sources(self):
        try:
            logging.info("Now start to fresh local sources(from /var/lib/kylin-system-updater to /etc/apt/sources.list)...")
            obj = self.bus.get_object('com.kylin.software.properties', '/com/kylin/software/properties')
            interface = dbus.Interface(obj, dbus_interface='com.kylin.software.properties.interface')
            interface.replaceUpgradeSource(timeout=0.5)
        except Exception as e:
            logging.warning(str(e))

    def _update_error_handler(self,error_code):
        self.now_working = InstallBackend.ACTION_DEFUALT_STATUS
        self.dbusController.UpdateDetectFinishedNext(False,[''],error_code)

    def _get_important_data(self):
        try:
            important_list = []
            if self.is_openkylin == False:
                with open(IMPORTANT_LIST_PATH, 'r') as f:
                    data = f.read()
                    important_list = data.split()
            else:
                important_list = [SYSTEM_UPDATE_GROUPS]

            #去除重复内容
            important_list = list(set(important_list))

            return important_list
        except Exception as e:
            logging.error(str(e))
            return []

    def _upload_dpkg_status(self):
        FILTER_SCRIPT="/usr/share/kylin-system-updater/scripts/filter-dpkg-status.sh"
        try:
            def calculate_md5(file_path):
                md5_hash = hashlib.md5()
                with open(file_path, "rb") as f:
                    # 逐块读取文件并更新MD5哈希对象
                    while chunk := f.read(4096):
                        md5_hash.update(chunk)

                return md5_hash.hexdigest()
            def _upload_data():
                result = subprocess.run([FILTER_SCRIPT],timeout=2)
                if result.returncode == 0:
                    logging.info("Streamling dpkg status information succeseed...")
                    self.collector.UpdateDpkgStatus()
                else:
                    logging.info("Streamling dpkg status information failed:...",str(result.stderr)+str(result.stdout))

            if self._is_upload_day():
                # 电量不足时禁止上传
                self._check_device_power(500)
                previous_md5 = self.configs_uncover.getWithDefault("DpkgStatusUpload","previous_md5","")
                status_dir = apt_pkg.config["Dir::State::status"]
                status_md5 = calculate_md5(status_dir)
                if previous_md5 != status_md5:
                    self.configs_uncover.setValue("DpkgStatusUpload","previous_md5",status_md5)
                    _upload_data()
            
            if self._force_upload_day():
                _upload_data()

        except Exception as e:
            logging.error(str(e))
        
        finally:
            return False

    def _is_upload_day(self):
        # type: () -> bool
        # 检查在一周的周几进行上传，返回True时进行触发上传
        upload_days = self.configs_cover.getlist("DpkgStatusUpload","upload_days")
        if not upload_days:
            return True
        # validate patch days
        today = date.today()
        # abbreviated localized dayname
        if today.strftime("%a") in upload_days:
            return True
        # full localized dayname
        if today.strftime("%A") in upload_days:
            return True
        # by number (Sun: 0, Mon: 1, ...)
        if today.strftime("%w") in upload_days:
            return True
        # today is not a patch day
        logging.info(
            "Skipping update check: today is %s,%s,%s but patch days are %s",
            today.strftime("%w"), today.strftime("%a"), today.strftime("%A"),
            ", ".join(upload_days))
        return False

    def _force_upload_day(self):
        # type: () -> bool
        def _update_days():
            random_days = random.randint(1, 28)
            self.configs_uncover.setValue("DpkgStatusUpload","force_upload_days",str(random_days))

        upload_days = self.configs_uncover.getWithDefault("DpkgStatusUpload","force_upload_days",0)
        if upload_days == 0:
            _update_days()
            return False
        else:
            today = date.today()
            if upload_days == today.day:
                _update_days()
                return True
            else:
                return False

    def umount_offline_source(self):
        if os.path.exists("/media/刻录"):
            logging.info("Staring umount /media/刻录")
            p = subprocess.run(["umount","-l","/media/刻录"],timeout=1)
            if p.returncode == 0:
                logging.info("umount offline source to success...")
                return False
            else:
                logging.info("umount offline source to failed...")
                return False

        if is_offline_source():
            if self.now_working != InstallBackend.ACTION_DEFUALT_STATUS:
                # logging.warning('In the process of other task(%d) and Start polling the umount...',self.now_working)
                return True
            try:
                logging.info("Staring umount /media/OfflineSource/")
                p = subprocess.run(["umount","-l","/media/OfflineSource/"],timeout=1)
                if p.returncode == 0:
                    logging.info("umount offline source to success...")
                    return False
                else:
                    logging.info("umount offline source to failed...")
                    return False
            except Exception as exc:
                logging.error(exc)
                return False
        else:
            logging.info("The offline source is not mounted, it does not need to be umounted...")
            return False

    def is_shutdown_install(self):
        if self.task_queue:
            self.last_install_mode = self.task_queue[-1].shutdown_install
            return self.last_install_mode
        else:
            return self.last_install_mode

    def require_shutdown_install(self):
        if self.task_queue and self.task_queue[-1].available and self.task_queue[-1].shutdown_install:
            logging.info("Requires a Shutdown installtion...")
            return 1
        else:
            logging.info("Do not need to perform a shutdown installtion...")
            return 0

    def _print_now_sources(self):
        SOURCES_LIST_FLAG = "sources.list"
        sources_path = apt_pkg.config["Dir::Etc::SourceList"]
        try:
            if sources_path != SOURCES_LIST_FLAG:
                with open(sources_path, 'r') as file:
                    file_content = file.read()
                    logging.info("Sources info:\n"+file_content)
        except Exception as e:
            logging.error(str(e))


    # private = True 代表 系统更新的私有配置 False：走本地系统的cache
    # fresh_cache True 代表强制刷新cache False：代表自动检查刷新cache 
    def refresh_cache(self,private=False,fresh_cache=False):
        SOURCES_LIST_FLAG = "sources.list"
        def _reload_private_options():
            if os.path.exists(APT_PRIVATE_PATH):
                apt_pkg.read_config_file(apt_pkg.config, APT_PRIVATE_PATH)
            apt_pkg.init_config()
        def _reload_standard_options():
            if os.path.exists(APT_STANDARD_PATH):
                apt_pkg.read_config_file(apt_pkg.config, APT_STANDARD_PATH)
            apt_pkg.init_config()

        try:
            sources_path = apt_pkg.config["Dir::Etc::SourceList"]
            if private:
                # 检查和加载的配置不同时 重新导入配置 以及加载cache
                if sources_path == SOURCES_LIST_FLAG:
                    _reload_private_options()
                    fresh_cache = True
            else: 
                if sources_path != SOURCES_LIST_FLAG:
                    _reload_standard_options()
                    fresh_cache = True
            if self.cache is None:
                self.cache = MyCache(None)
            elif fresh_cache:
                self.cache.open(None)
                self.cache._initDepCache()
                logging.info("Finished: refreshing Cache for System Update Successfully...")
        except AssertionError:
            raise UpdateBaseError(ERROR_SOFTWARE_INDEX_RROKEN)
        except SystemError as e:
            logging.error(str(e))
            raise UpdateBaseError(ERROR_NOT_INIT_PACKAGESINFIO)
        return self.cache

    def _setup_dbus(self,bus):
        # check if there is another g-a-i already and if not setup one
        # listening on dbus
        try:
            bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                            bus,do_not_queue=True)
            logging.info("Initiate dbus success ...")
            UpdateManagerDbusControllerUtils(self,bus,bus_name)
            return  UpdateManagerDbusController(self,bus,bus_name)
        except dbus.exceptions.NameExistsException:
            if self.options.replace is False:
                logging.critical("Another daemon is already running")
                sys.exit(1)
            logging.warning("Replacing already running daemon")
            
            retry_reboot_times = 0
            the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                            UPDATER_DBUS_PATH)
            the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                timeout=300)
            time.sleep(1)
            while True:
                retry_reboot_times = retry_reboot_times + 1
                #当重试次数超过5次时退出程序
                if retry_reboot_times > 5:
                    logging.critical("Reboot backend is Failed...")
                    sys.exit(1)
                try:
                    bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                                    bus,
                                                    do_not_queue=True)
                    logging.warning("Replacing already running daemon to Success...")
                    return  UpdateManagerDbusController(self,bus,bus_name)
                except dbus.exceptions.NameExistsException:
                    the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                                    UPDATER_DBUS_PATH)
                    the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                        timeout=300)
                    logging.error("Dbus has not withdrawn and retry reboot times:%d...",retry_reboot_times)
                    time.sleep(1)
    
    def _setup_dbus_utils(self,bus):
        # check if there is another g-a-i already and if not setup one
        # listening on dbus
        try:
            bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                            bus,
                                            do_not_queue=True)
            return  UpdateManagerDbusControllerUtils(self,bus,bus_name)
        except dbus.exceptions.NameExistsException:
            if self.options.replace is False:
                logging.critical("Another daemon is already running")
                sys.exit(1)
            logging.warning("Replacing already running daemon")
            
            retry_reboot_times = 0
            the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                            UPDATER_DBUS_PATH)
            the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                timeout=300)
            time.sleep(1)
            while True:
                retry_reboot_times = retry_reboot_times + 1
                #当重试次数超过5次时退出程序
                if retry_reboot_times > 5:
                    logging.critical("Reboot backend is Failed...")
                    sys.exit(1)
                try:
                    bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                                    bus,
                                                    do_not_queue=True)
                    logging.warning("Replacing already running daemon to Success...")
                    return  UpdateManagerDbusController(self,bus,bus_name)
                except dbus.exceptions.NameExistsException:
                    the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                                    UPDATER_DBUS_PATH)
                    the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                        timeout=300)
                    logging.error("Dbus has not withdrawn and retry reboot times:%d...",retry_reboot_times)
                    time.sleep(1)

    def start_back_upgrade(self, pkglist, mode=InstallBackend.MODE_OTHER):
        try:
            install_backend = get_backend(self, InstallBackend.ACTION_BACKGROUND_UPGRADE, mode)
            install_backend.start_alone(push_content = pkglist)
        except Exception as e:
            logging.error(str(e))

    def start_back_update_cache(self, action= InstallBackend.ACTION_BACKGROUND_UPGRADE, mode=InstallBackend.MODE_OTHER):
        try:
            install_backend = get_backend(self, action, mode)
            install_backend.start_alone()
        except Exception as e:
            logging.error(str(e))

    def start_deb_resolver(self, pkglist = []):
        try:
            install_backend = get_backend(self, InstallBackend.ACTION_INSTALL_DEB_RESOLVER)
            install_backend.start_alone(push_content = pkglist)
        except Exception as e:
            logging.error(str(e))

    # 进行本地deb包安装的操作
    # _check_local_dep : 是否查询本地依赖
    # _auto_satisfy    : 是否通过网络下载依赖
    def start_deb_install(self, deb_path = "", _check_local_dep = False, _auto_satisfy = False, source = '', sender=None):
        header = ''
        desc = ''
        absolute_path, debname = os.path.split(deb_path)
        self.sqlite3_server.deb_metadata.update({"current_install_debfile":deb_path})
        self.sqlite3_server.deb_metadata.update({"absolute_path":absolute_path})
        
        self.deb_obj = {}
        UpdateMsg = {}
        try:
            if not os.path.isfile(deb_path):
                logging.info("No such file or directory: %s.",deb_path)
                self.dbusController.InstalldebFinished(False,'No such file or directory .','')
                self.dbusController.InstalldebFinishedWithErrCode(False,'No such file or directory .','',get_error_num_from_enum(ERROR_READ_LOCAL_DEB))
                return
            # 验签提权
            sender_name = get_proc_from_dbus_name(sender)
            caller = get_caller_from_enum(sender_name)
            caller_trans = get_source_name_from_enum(sender_name)
            self.sqlite3_server.deb_metadata.update({"source":str(sender_name)})

            def _run_install_handler():
                self.deb_obj.update({"debname":str(debname)})
                self.deb_obj.update({"old_version":""})
                self.deb_obj.update({"source":str(caller)})
                UpdateMsg.update({"source":str(self.deb_obj.get("source","kylin-system-updater"))})

                try:
                    self.refresh_cache(False,True)
                except Exception as e:
                    logging.warning("refresh_cache error, install with dpkg")
                    _success,header,desc = self._dpkg_install_pkgs(deb_path)
                    if _success == True:
                        self.dbusController.InstalldebFinished(True,header,desc)
                        self.dbusController.InstalldebFinishedWithErrCode(True,header,desc,get_error_num_from_enum(SUCCESS_INSTALL_DEB))
                    else:
                        self.dbusController.InstalldebFinished(False,header,desc)
                        self.dbusController.InstalldebFinishedWithErrCode(False,header,desc,get_error_num_from_enum(ERROR_INSTALL_DEB_DPKG))
                    return

                if self.cache is None:
                    deb_cache = MyCache(None)
                else:
                    deb_cache = self.cache

                ins = self._suit_install_mode(deb_path, deb_cache)
                UpdateMsg.update({"appname":str(self.debName)})
                self.sqlite3_server.deb_metadata.update({"debname":str(self.debName)})
                self.sqlite3_server.deb_metadata['deblist'].append(str(self.debName))
                UpdateMsg.update({"new_version":str(self.debVersion)})

                status,error_string = ApplicationSecInstall(str(self.debName))
                if not status:
                    self.dbusController.InstalldebFinished(False,error_string,'')
                    self.dbusController.InstalldebFinishedWithErrCode(False,error_string,'',get_error_num_from_enum(ERROR_INSTALL_DEB_SECURITY))
                    return 

                if self._is_dpkgJournalDirty:
                    _success,header,desc = self._dpkg_install_pkgs(deb_path)
                    if _success == True:
                        self.dbusController.InstalldebFinished(True,header,desc)
                        self.dbusController.InstalldebFinishedWithErrCode(True,header,desc,get_error_num_from_enum(SUCCESS_INSTALL_DEB))
                    else:
                        self.dbusController.InstalldebFinished(False,header,desc)
                        self.dbusController.InstalldebFinishedWithErrCode(False,header,desc,get_error_num_from_enum(ERROR_INSTALL_DEB_DPKG))
                elif self._is_broken > 0 or not self.cacheSatisfy:
                # 走 dpkg 安装流程，说明本地apt环境已经损坏 or dep not satisfied
                    dep_satisfy, header, desc = self._deb_install(deb_cache, deb_path, _check_local_dep, ins)
                    if dep_satisfy:
                        self.dbusController.InstalldebFinished(True, header, desc)
                        self.dbusController.InstalldebFinishedWithErrCode(True, header, desc,get_error_num_from_enum(SUCCESS_INSTALL_DEB))
                        UpdateMsg.update({"status":"success"})
                        UpdateMsg.update({"errorCode":" "})
                    else:
                        self.dbusController.InstalldebFinished(False, header, desc)
                        self.dbusController.InstalldebFinishedWithErrCode(False, header, desc,get_error_num_from_enum(ERROR_INSTALL_DEB_DPKG))
                        UpdateMsg.update({"status":"failed"})
                        UpdateMsg.update({"errorCode":str(header+"， "+desc)})
                    #dpkg发送数据
                    if self.configs_uncover.getWithDefault("SystemStatus", "upload_installer_log", False) == True:
                        UpdateMsg.update({"old_version":str(self.deb_obj.get("old_version","None"))})
                        self.collector.Upgrade_Process_Msg(InstallBackend.ACTION_INSTALL_DEB, UpdateMsg.copy())
                        self.deb_obj = {}
                else:
                # apt 安装流程
                    dep_satisfy, header, desc = self._attempt_depends(deb_cache, deb_path, _check_local_dep,_auto_satisfy, ins)
                    if dep_satisfy:
                        install_backend = get_backend(self, InstallBackend.ACTION_INSTALL_DEB)
                        logging.info("source name: %s.", source)
                        install_backend.start_alone(push_content=deb_path,_is_install=_auto_satisfy,caller=source)
                    else:
                        self.dbusController.InstalldebFinished(False, header, desc)
                        self.dbusController.InstalldebFinishedWithErrCode(False, header, desc,get_error_num_from_enum(ERROR_INSTALL_DEB_DEPEND))

            def on_authorization_success(result):
                logging.info("Authorization success ...")

                self.sqlite3_server.deb_policy_keep = True
                def _check_deb_policy():
                    if self.sqlite3_server.deb_policy_timestamp % 10 == 0:
                        logging.info("Checking for deb policy timeout(%d)...",self.sqlite3_server.deb_policy_timestamp)
                    if (self.sqlite3_server.deb_policy_timestamp <= 0):
                        logging.warning("Deb policy timeout")
                        self.sqlite3_server.deb_policy_keep = False
                        return False
                    else:
                        self.sqlite3_server.deb_policy_timestamp = self.sqlite3_server.deb_policy_timestamp - 1
                    return True

                from gi.repository import GLib
                self.sqlite3_server.deb_policy_timestamp = 60 * 5
                GLib.timeout_add_seconds(1,_check_deb_policy)

                _run_install_handler()

            def on_authorization_failure(error):
                """Convert the DeferredException to a normal exception."""
                try:
                    logging.info("Authorization failed ...")

                    self.sqlite3_server.deb_policy_keep = False
                    self.sqlite3_server.deb_policy_timestamp = 0

                    error.raise_exception()
                except Exception as error:
                    logging.error(error)
                    self.dbusController.InstalldebFinished(False,_("Cancel authentication."),'')
                    self.dbusController.InstalldebFinishedWithErrCode(False,_("Cancel authentication."),'',get_error_num_from_enum(ERROR_INSTALL_DEB_VERIFY))

            def _policykit_authority_handler(install_handler = None, authority_failed_handler = None):
                if not self.sqlite3_server.deb_policy_keep:
                    deferred = PolicyKit_Authority(caller_trans+_(" requires authentication to install software packages."),
                                                sender = sender, InstPolicy = True,
                                                authentication = self.configs_uncover.getWithDefault(
                                                                                        "InstallAndPurge",
                                                                                        "install_authority",True),
                                                source=source)

                    deferred.add_callback(on_authorization_success)
                    deferred.add_errback(on_authorization_failure)
                else:
                    self.sqlite3_server.deb_policy_timestamp = 60 * 5
                    logging.info("Deb policy keep, ignore...")

                    _run_install_handler()
        
            if not deb_verify(deb_path): #验签失败，提权
                _policykit_authority_handler()
            else:
                _run_install_handler()

        except UpdateBaseError as excep:
            self.dbusController.InstalldebFinished(False,excep.header,excep.desc)
            self.dbusController.InstalldebFinishedWithErrCode(False,excep.header,excep.desc,get_error_num_from_enum(excep.code))
            UpdateMsg.update({"old_version":str(self.deb_obj.get("old_version","None"))})
            UpdateMsg.update({"errorCode":str(excep.header+"， "+excep.desc)})
            UpdateMsg.update({"status":"failed"})
            self.collector.Upgrade_Process_Msg(InstallBackend.ACTION_INSTALL_DEB, UpdateMsg.copy())
        except Exception as e:
            logging.info(str(e))
            tbk = traceback.format_exc()
            logging.error(str(tbk))
            self.dbusController.InstalldebFinished(False, str(e), desc)
            self.dbusController.InstalldebFinishedWithErrCode(False,str(e),desc,get_error_num_from_enum(ERROR_INSTALL_DEB_OTHER))
            UpdateMsg.update({"old_version":str(self.deb_obj.get("old_version","None"))})
            UpdateMsg.update({"errorCode":str(e)})
            UpdateMsg.update({"status":"failed"})
            self.collector.Upgrade_Process_Msg(InstallBackend.ACTION_INSTALL_DEB, UpdateMsg.copy())

    def _system_reboot_shutdown(self,status="reboot",prohibit_shutdown=False):
        if status == "shutdown":
            args = ["/sbin/shutdown","-h","now"]
        else:
            args = ["/sbin/shutdown","-r","now"]

        if prohibit_shutdown:
            logging.info("The simulate mode and run system(%s) command(%r)...",status,args)
        else:
            logging.info("The device will be %s by run system command(%r)...",status,args)
            p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
            if p.returncode != 0:
                logging.info("The device has (%s) Failed...",status)
                return False
            
    #进行删除的操作
    def start_purge_pkgs(self,pkgs_list,sender=None):
        try:
            sender_name = get_proc_from_dbus_name(sender)
            caller = get_caller_from_enum(sender_name)
            caller_trans = get_source_name_from_enum(sender_name)

            def _run_purge_handler():
                broken_count = 0
                use_dpkg_pruge = False
                try:
                    # 是否有破损的包
                    self.refresh_cache(False,True)
                    if self.cache is None:
                        deb_cache = MyCache(None)
                    else:
                        deb_cache = self.cache
                    broken_count = deb_cache._depcache.broken_count
                except Exception as e:
                    use_dpkg_pruge = True

                if use_dpkg_pruge or broken_count > 0:
                    _success,header,desc = self._dpkg_purge_pkgs(pkgs_list)
                    if _success == True:
                        logging.info(header)
                        self.dbusController.PurgePackagesFinished(_success,'',desc)
                        self.dbusController.PurgePackagesFinishedWithPkgname(_success,'',desc,",".join(pkgs_list))
                        self.dbusController.PurgePackagesFinishedWithErrCode(_success,'',desc,",".join(pkgs_list),get_error_num_from_enum(SUCCESS_PURGE_PKG))
                    else:
                        self.dbusController.PurgePackagesFinished(_success,header,desc)
                        self.dbusController.PurgePackagesFinishedWithPkgname(_success,header,desc,",".join(pkgs_list))
                        self.dbusController.PurgePackagesFinishedWithErrCode(_success,header,desc,",".join(pkgs_list),get_error_num_from_enum(ERROR_PURGE_PKG_DPKG))
                else:
                    purge_backend = get_backend(self, InstallBackend.ACTION_REMOVE_PACKAGES)
                    purge_backend.start_alone(push_content = pkgs_list)

            def on_authorization_success(result):
                logging.info("Authorization success ...")

                self.sqlite3_server.purge_policy_keep = True
                def _check_purge_policy():
                    if self.sqlite3_server.purge_policy_timestamp % 10 == 0:
                        logging.info("Checking for deb policy timeout(%d)...",self.sqlite3_server.purge_policy_timestamp)
                    if (self.sqlite3_server.purge_policy_timestamp <= 0):
                        logging.warning("Deb policy timeout")
                        self.sqlite3_server.purge_policy_keep = False
                        return False
                    else:
                        self.sqlite3_server.purge_policy_timestamp = self.sqlite3_server.purge_policy_timestamp - 1
                    return True

                from gi.repository import GLib
                self.sqlite3_server.purge_policy_timestamp = 60 * 5
                GLib.timeout_add_seconds(1,_check_purge_policy)

                _run_purge_handler()

            def on_authorization_failure(error):
                """Convert the DeferredException to a normal exception."""
                try:
                    logging.info("Authorization failed ...")

                    self.sqlite3_server.purge_policy_keep = False
                    self.sqlite3_server.purge_policy_timestamp = 0

                    error.raise_exception()
                except Exception as error:
                    logging.error(error)
                    self.dbusController.PurgePackagesFinished(False,_("Cancel authentication."),'')
                    self.dbusController.PurgePackagesFinishedWithPkgname(False,_("Cancel authentication."),'',",".join(pkgs_list))
                    self.dbusController.PurgePackagesFinishedWithErrCode(False,_("Cancel authentication."),'',",".join(pkgs_list),get_error_num_from_enum(ERROR_PURGE_PKG_VERIFY))

            def _policykit_authority_handler(install_handler = None, authority_failed_handler = None):
                if not self.sqlite3_server.purge_policy_keep:
                    deferred = PolicyKit_Authority(
                                    get_source_name_from_enum(sender_name)+_(" requires authentication to uninstall software packages."),
                                    sender = sender, InstPolicy = False, 
                                    authentication = self.configs_uncover.getWithDefault("InstallAndPurge","purge_authority",False), 
                                    source=sender_name)

                    deferred.add_callback(on_authorization_success)
                    deferred.add_errback(on_authorization_failure)
                else:
                    self.sqlite3_server.purge_policy_timestamp = 60 * 5
                    logging.info("Purge policy keep, ignore...")

                    _run_purge_handler()

            _policykit_authority_handler()

        except Exception as e:
            logging.error(e)
            tbk = traceback.format_exc()
            logging.error(str(tbk))

    def _dpkg_purge_pkgs(self,pkgs_list):
        success = False
        pkg_list_str = " ".join(pkgs_list)
        args = ["dpkg", "-r"]
        args.extend([pkg_list_str])

        p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
        success = p.returncode == 0
        return success,p.stdout,''

    def _dpkg_install_pkgs(self,deb_path):
        logging.info("start dpkg install: %s", deb_path)
        success = False
        args = ["dpkg", "-i"]
        args.extend([deb_path])

        p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
        success = p.returncode == 0
        return success,p.stdout,''

    # 是否查找本地依赖
    def _attempt_depends(self,deb_cache, deb_path,_check_local_dep,_auto_satisfy, _install):  
        depends_list = []
        depends_pkg = []
        satisfy_list = []
        depends_count = 0
        _local_satisfy = False
        error_string = ''
        error_desc = ''
        absolute_path, debname = os.path.split(deb_path)
        # 依赖不满足的情况
        if "_" in debname and _install:
            if str(debname.split("_")[0]) in _install:
                _install.remove(str(debname.split("_")[0]))
        if len(_install) > 0:
            if _check_local_dep: #查找本地
                # 需要查找本地依赖
                if len(_install) > 0:
                    for pkg in deb_cache:
                        if pkg.marked_install and pkg.name != str(debname.split("_")[0]):
                            depends_pkg.append(pkg)
                        elif pkg.marked_upgrade and pkg.name != str(debname.split("_")[0]):
                            depends_pkg.append(pkg)
                if len(depends_pkg)>0: #查找本地deb包
                    depends_list = [debfile for debfile in os.listdir(absolute_path) if debfile.endswith(".deb")]
                    for depends in depends_pkg:
                        for debfile in depends_list:
                            if depends.name in debfile and depends.candidate.version in debfile:
                                depends_count += 1
                                satisfy_list.append(debfile)
                    if depends_count < len(depends_pkg):
                        #本地依赖不满足 判断源下载
                        if _auto_satisfy:
                            _local_satisfy = True
                        elif not _auto_satisfy:
                            _local_satisfy = False
                            error_string = str(debname.split("_")[0])+_("dependency is not satisfied")
                            error_desc   = ",".join(_install)
                            logging.error(error_string+ error_desc)
                    else:
                        #将应用包与依赖包拷贝至archive目录安装
                        try:
                            if debname not in satisfy_list:
                                satisfy_list.append(debname)
                            for satisfy in satisfy_list:
                                shutil.copy(os.path.join(absolute_path,satisfy),"/var/cache/apt/archives/")
                                logging.info("move debpkg: %s",satisfy)
                            _local_satisfy = True
                        except Exception as e:
                            logging.info(str(e))
                return _local_satisfy,error_string,error_desc
            elif not _check_local_dep and _auto_satisfy:
                _local_satisfy = True
                if _install:
                    error_string = str(debname.split("_")[0])+_("dependency is not satisfied will download")
                    error_desc   = ",".join(_install)
                    logging.error(error_string+error_desc)
                    return _local_satisfy,error_string,error_desc
            elif not _check_local_dep and not _auto_satisfy:
                _local_satisfy = False
                if _install:
                    error_string = str(debname.split("_")[0])+_("dependency is not satisfied")
                    error_desc   = ",".join(_install)
                    logging.error(error_string+error_desc)
                    return _local_satisfy,error_string,error_desc
        # 依赖满足
        else:
            _local_satisfy = True
            error_string = ''
            error_desc=''
            return _local_satisfy,error_string,error_desc
    
    def _deb_install(self, deb_cache, deb_path, _check_local_dep, _install):
        local_depends_deb_list = []
        satisfy_list = []
        noSatisfyList = []
        self.depends_count = 0
        error_string = ''
        header = ''
        desc = ''
        absolute_path, debname = os.path.split(deb_path)
        debPackage = DebPackage(deb_path, deb_cache)
        self.absolute_path = absolute_path

        if _check_local_dep: # 需要查找本地依赖
            if len(self.noSatisfyList) > 0:
                for nS in self.noSatisfyList:
                    if str(nS[0][0]) not in noSatisfyList:
                        noSatisfyList.append(str(nS[0][0]))

            if len(noSatisfyList) > 0: #查找本地deb包
                local_depends_deb_list = [debfile for debfile in os.listdir(absolute_path) if debfile.endswith(".deb")]

                for nS in self.noSatisfyList:
                    for debfile in local_depends_deb_list:
                        # if "%3a" in debfile:
                        #     debfile=debfile.replace("%3a",":")
                        dep = DebPackage(os.path.join(absolute_path,debfile), deb_cache)
                        
                        if str(nS[0][0]) == dep.pkgname:# and debfile not in satisfy_list:
                            is_resolve = self._depend_resolver(dep, debfile, deb_cache, local_depends_deb_list, satisfy_list)
                            if is_resolve and debfile not in satisfy_list:
                                self.depends_count += 1
                                satisfy_list.append(debfile)

                if self.depends_count < len(noSatisfyList): #本地依赖不满足
                    error_string = str(debPackage.pkgname)+_("dependency is not satisfied")+", ".join(noSatisfyList)
                    logging.error(error_string)
                    header = error_string
                    return False,header, desc
                else: #安装依赖包
                    if debname not in satisfy_list:
                        satisfy_list.append(debname)
                    for satisfy in satisfy_list:
                        try:
                            deb = DebPackage(os.path.join(absolute_path,satisfy))
                            # iprogress = LogInstallProgress(os.path.join(absolute_path,satisfy))
                            ret = deb.install() #install_progress=iprogress
                            if ret != 0:
                                return False, "软件包安装失败!", desc
                        except Exception as e:
                            logging.error(e)
                            return False, str(e), desc
                    return True, header, desc
            else:
                try:
                    deb = DebPackage(deb_path)
                    # iprogress = LogInstallProgress(os.path.join(absolute_path,satisfy))
                    ret = deb.install() #install_progress=iprogress
                    if ret != 0:
                        return False, "软件包安装失败!", desc
                except Exception as e:
                    logging.error(e)
                    return False, str(e), desc
                return True, header, desc
        elif not _check_local_dep:
            error_string = str(debPackage.pkgname)+_("dependency is not satisfied")+":".join(_install)
            logging.error(error_string)
            header = error_string
            return False,header, desc

    def _depend_resolver(self, deb, debfile, deb_cache, local_depends_deb_list, satisfy_list):
        depends_cnt = 0
        is_resolve = False
        depends = deb.depends
        noSatisfyList = self._gen_noSatisfyList(depends, deb_cache)
        if len(noSatisfyList) > 0:
            for nS in noSatisfyList:
                for debfile in local_depends_deb_list:
                    if "%3a" in debfile:
                        debfile=debfile.replace("%3a",":")
                    dep = DebPackage(os.path.join(self.absolute_path,debfile), deb_cache)

                    if str(nS[0][0]) == dep.pkgname:
                        is_resolve = self._depend_resolver(dep, debfile, deb_cache, local_depends_deb_list, satisfy_list)
                        if is_resolve and debfile not in satisfy_list:
                            depends_cnt += 1
                            satisfy_list.append(debfile)
            if depends_cnt < len(noSatisfyList):
                is_resolve = False
        else:
                is_resolve = True
        return is_resolve

    def _suit_install_mode(self, deb_path, deb_cache):
        install = []
        self._is_broken = False
        self.cacheSatisfy = False
        self.debVersion = " "
        self.debName = deb_path
        self._is_dpkgJournalDirty = False

        if self.cache is None:
            return install

        # 检查本地破损
        try:
            broken_count = deb_cache._depcache.broken_count
            if deb_cache.dpkgJournalDirty():
                logging.info("dpkgJournalDirty True")
                self._is_dpkgJournalDirty = True
            logging.info("Install deb_package: check broken: %d. ", broken_count)
            self.debVersion = 0
            debPackage = DebPackage(deb_path, deb_cache)
            debPackage.check(allow_downgrade=True)
            self.debVersion = debPackage._sections["Version"]
            self.debName = debPackage.pkgname
            logging.info("Install deb_package: required changes")
            (install, remove, unauth) = debPackage.required_changes    # need in cach
            if broken_count > 0:
                self._is_broken = True
            else :
                self._is_broken = False
            if debPackage.pkgname in deb_cache:
                pkg = deb_cache[debPackage.pkgname]
                if pkg.is_installed:
                    self.deb_obj.update({"old_version":str(pkg.installed.version)})
        except apt_pkg.Error as e:
            logging.error(str(e))
            if "E:read" in str(e):
                raise UpdateBaseError(ERROR_READ_LOCAL_DEB)
            else:
                raise UpdateBaseError(ERROR_LOCAL_DEB_FORMAT)
        except Exception as e:
            logging.error(str(e))
            raise UpdateBaseError(ERROR_INSTALL_DEB_BASE)
        
        # 不满足的依赖列表
        depends = debPackage.depends
        self.noSatisfyList = self._gen_noSatisfyList(depends, deb_cache)
        _list = []
        # cache是否满足
        if len(install) == 0 and len(self.noSatisfyList) == 0:
            self.cacheSatisfy = True
        else:
            for ns in self.noSatisfyList:
                for or_group in ns:
                    for pkg in install:
                        if pkg == or_group[0]:
                            _list.append(ns)
            if len(_list) == len(self.noSatisfyList):
                self.cacheSatisfy = True
            else:
                self.cacheSatisfy = False
        logging.info("Cache satisfy is %r.",self.cacheSatisfy)
        return install
    
    def _gen_noSatisfyList(self, depends, deb_cache):
        _noSatisfyList = []
        _group_satify=False
        providers = []
        for or_group in depends:
            for deb_info in or_group:
                debname,ver,oper = deb_info
                if ":" in debname:
                    debname=debname.split(":")[0]
                try:
                    if _group_satify:
                        logging.warning("Group depend satify, ignore %s.", debname)
                        continue
                    if debname not in deb_cache:
                        if  deb_cache.is_virtual_package(debname):
                            logging.warning("The cache has no package named '%s', this is a virtual package...", debname)
                            providers = deb_cache.get_providing_packages(debname)
                        # hear!
                        if len(providers) < 1:
                            logging.error("Can not found depend %s.", debname)
                            continue
                        depname = providers[0].name
                        pkg = deb_cache[depname]
                        cand = deb_cache._depcache.get_candidate_ver(pkg._pkg)
                        logging.warning("Found virtual package %s, providers name: %s.", debname,depname)
                        if not cand:
                            _group_satify = False
                        if not apt_pkg.check_dep(cand.ver_str, oper, ver):
                            _group_satify = False
                        else:
                            _group_satify = True
                    else:
                        pkg = deb_cache[debname]
                        if (apt_pkg.check_dep(pkg.candidate.source_version, oper, ver)) or (pkg.installed and apt_pkg.check_dep(pkg.installed.version, oper, ver)) or (not pkg.installed and apt_pkg.check_dep(pkg.candidate.version, oper, ver)) and _group_satify == False:
                            # if not apt_pkg.check_dep(cand, oper, ver): #no candidate
                            _group_satify = True
                except Exception as e:
                    logging.error(str(e))
            if _group_satify == False and or_group not in _noSatisfyList:
                _noSatisfyList.append(or_group)
            _group_satify = False
        return _noSatisfyList


class AptP2pConfigManager():
    APT_P2P_FILE = "/etc/apt-p2p/apt-p2p.conf"
    HEADER_DSC = "apt-p2p config(/etc/apt-p2p/apt-p2p.conf) is not exists..."
    def __init__(self):
        if os.path.exists(self.APT_P2P_FILE):
            self.p2pConfigs = UpgradeConfig(datadir = "/etc/apt-p2p/", name = "apt-p2p.conf")
        else:
            self.p2pConfigs = None
    
    def get_bootstrap(self):
        if self.p2pConfigs == None:
            return self.HEADER_DSC

        return self.p2pConfigs.getWithDefault("apt_p2p_Khashmir", "BOOTSTRAP", "Failed")

    def set_bootstrap(self,value):
        if self.p2pConfigs == None:
            return self.HEADER_DSC
        self.p2pConfigs.setValue("apt_p2p_Khashmir","BOOTSTRAP",str(value),True)